from typing import Union, overload
from decimal import Decimal, ROUND_HALF_UP
from babel.numbers import get_currency_precision
import warnings

Dint = Union[Decimal, int]


class Currency:
    """Handles money per specified currency"""

    __slots__ = ('amount', 'currency')

    def __init__(self, amount: Dint, currency: str) -> None:
        if isinstance(amount, float):
            warnings.warn(
                SyntaxWarning(  # pragma: no cover
                    'float value detected. Please use Decimal instead.'
                ),
                stacklevel=2
            )
        self.amount = Decimal(amount)
        self.currency = currency.upper() or 'USD'

    def __str__(self) -> str:
        return f'{str(self.amount)} {self.currency}'

    def __bool__(self) -> bool:
        return bool(self.amount)

    @overload
    def __truediv__(self, other: Dint) -> 'Currency':
        ...  # no pragma cover

    @overload
    def __truediv__(self, other: 'Currency') -> 'Dint':
        ...  # no pragma cover

    def __truediv__(self, other):
        if isinstance(other, Currency):
            if self.currency != other.currency:
                raise ValueError(
                    f'cannot div {self.currency} from {other.currency}'
                )
            return self.amount / other.amount
        try:
            amount = self.amount / other
        except TypeError:
            return NotImplemented
        return Currency(
            amount,
            self.currency
        )

    def __add__(self, other: 'Currency') -> 'Currency':
        if isinstance(other, Currency):
            if self.currency != other.currency:
                raise ValueError(
                    f'cannot add {self.currency} to {other.currency}'
                )
            return Currency(
                self.amount + other.amount,
                self.currency
            )
        return NotImplemented

    def __sub__(self, other: 'Currency') -> 'Currency':
        """negative currency situation needs to be handled externally"""
        if isinstance(other, Currency):
            if self.currency != other.currency:
                raise ValueError(
                    f'cannot subtract {self.currency} from {other.currency}'
                )
            amount = self.amount - other.amount
            return Currency(
                amount,
                self.currency
            )
        return NotImplemented

    def __mul__(self, other: Dint) -> 'Currency':
        try:
            amount = self.amount * other
        except TypeError:
            return NotImplemented
        return Currency(amount, self.currency)

    def __rmul__(self, other: Dint) -> 'Currency':
        return self * other

    def __lt__(self, other: 'Currency') -> bool:
        if isinstance(other, Currency):
            if self.currency != other.currency:
                raise ValueError(
                    f'cannot compare {self.currency} with {other.currency}'
                )
            return self.amount < other.amount
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Currency):
            return (
                    self.amount == other.amount and
                    self.currency == other.currency)
        return False

    def __gt__(self, other: 'Currency') -> bool:
        if isinstance(other, Currency):
            if self.currency != other.currency:
                raise ValueError(
                    f'cannot compare {self.currency} with {other.currency}'
                )
            return self.amount > other.amount
        return NotImplemented

    def __le__(self, other: 'Currency') -> bool:
        if self == other:
            return True
        return self < other

    def __ge__(self, other: 'Currency') -> bool:
        if self == other:
            return True
        return self > other

    def quantize(self, exp=None, rounding=None) -> 'Currency':
        """
        :param exp: fractional part of currency. Like cents in dollar. It can be set to any arbitrary amount.
        :param rounding: ROUND_DOWN, ROUND_UP, ROUND_HALF_UP
        :return: Currency object
        The quantize() method rounds a number to a fixed exponent.
        This method is useful for monetary applications that
        often round results to a fixed number of places
        Example: (10.001, 'USD').quantize(exp='.01', rounding=ROUND_DOWN)
        results in (10.00, 'USD')
        """
        if exp is None:
            exp = Decimal('0.1') ** get_currency_precision(self.currency)
        else:
            exp = Decimal(exp)
        if rounding is None:
            rounding = ROUND_HALF_UP
        return Currency(
            self.amount.quantize(exp=exp, rounding=rounding),
            self.currency
        )





from decimal import Decimal, ROUND_DOWN
from typing import TypeVar, Union

from tekmoney.currency import Currency
from tekmoney.currency_range import CurrencyRange
from tekmoney.currency_tax import CurrencyWithTax
from tekmoney.currency_range_tax import CurrencyRangeTax

Dint = Union[Decimal, int]

T = TypeVar('T', Currency, CurrencyRange, CurrencyWithTax, CurrencyRangeTax)


def fixed_discount(base: T, discount: Currency) -> T:
    """Apply a fixed discount to any currency type."""
    if isinstance(base, CurrencyRange):
        return CurrencyRange(
            fixed_discount(base.start, discount),
            fixed_discount(base.stop, discount))
    if isinstance(base, CurrencyRangeTax):
        return CurrencyRangeTax(
            fixed_discount(base.start, discount),
            fixed_discount(base.stop, discount))
    if isinstance(base, CurrencyWithTax):
        return CurrencyWithTax(
            net=fixed_discount(base.net, discount),
            gross=fixed_discount(base.gross, discount))
    if isinstance(base, Currency):
        return max(base - discount, Currency(0, base.currency))
    raise TypeError('Unknown base for fixed_discount: %r' % (base,))


def fractional_discount(base: T, fraction: Decimal, *, from_gross=True) -> T:
    """Apply a fractional discount based on either gross or net amount."""
    if isinstance(base, CurrencyRange):
        return CurrencyRange(
            fractional_discount(base.start, fraction, from_gross=from_gross),
            fractional_discount(base.stop, fraction, from_gross=from_gross))
    if isinstance(base, CurrencyRangeTax):
        return CurrencyRangeTax(
            fractional_discount(base.start, fraction, from_gross=from_gross),
            fractional_discount(base.stop, fraction, from_gross=from_gross))
    if isinstance(base, CurrencyWithTax):
        if from_gross:
            discount = (base.gross * fraction).quantize(rounding=ROUND_DOWN)
        else:
            discount = (base.net * fraction).quantize(rounding=ROUND_DOWN)
        return fixed_discount(base, discount)
    if isinstance(base, Currency):
        discount = (base * fraction).quantize(rounding=ROUND_DOWN)
        return fixed_discount(base, discount)
    raise TypeError('Unknown base for fractional_discount: %r' % (base,))


def percentage_discount(base: T, percentage: Dint, *, from_gross=True) -> T:
    """Apply a percentage discount based on either gross or net amount."""
    factor = Decimal(percentage) / 100
    return fractional_discount(base, factor, from_gross=from_gross)



from decimal import Decimal

import pytest

from tekmoney.currency import Currency
from tekmoney.currency_tax import CurrencyWithTax
from tekmoney.currency_range import CurrencyRange
from tekmoney.currency_range_tax import CurrencyRangeTax
from tekmoney.flat_tax import flat_tax


def test_application():
    result = flat_tax(CurrencyWithTax(Currency(10, 'BTC'), Currency(10, 'BTC')), 1)
    assert result.net == Currency(10, 'BTC')
    assert result.gross == Currency(20, 'BTC')
    result = flat_tax(Currency(100, 'BTC'), Decimal('0.5'))
    assert result.net == Currency(100, 'BTC')
    assert result.gross == Currency(150, 'BTC')
    with pytest.raises(TypeError):
        flat_tax(1, 1)


def test_tax_from_gross():
    result = flat_tax(
        CurrencyWithTax(Currency(120, 'USD'), Currency(120, 'USD')),
        Decimal('0.2'), keep_gross=True)
    assert result.net == Currency(100, 'USD')
    assert result.gross == Currency(120, 'USD')
    result = flat_tax(Currency(150, 'BTC'), Decimal('0.5'), keep_gross=True)
    assert result.net == Currency(100, 'BTC')
    assert result.gross == Currency(150, 'BTC')


def test_range():
    price_range = CurrencyRange(Currency(10, 'BTC'), Currency(20, 'BTC'))
    result = flat_tax(price_range, 1)
    assert result.start == CurrencyWithTax(Currency(10, 'BTC'), Currency(20, 'BTC'))
    assert result.stop == CurrencyWithTax(Currency(20, 'BTC'), Currency(40, 'BTC'))
    price_range = CurrencyRangeTax(
        CurrencyWithTax(Currency(10, 'BTC'), Currency(10, 'BTC')),
        CurrencyWithTax(Currency(20, 'BTC'), Currency(20, 'BTC')))
    result = flat_tax(price_range, 1)
    assert result.start == CurrencyWithTax(Currency(10, 'BTC'), Currency(20, 'BTC'))
    assert result.stop == CurrencyWithTax(Currency(20, 'BTC'), Currency(40, 'BTC'))



from functools import partial
from decimal import Decimal
import pytest

from tekmoney.currency import Currency
from tekmoney.currency_range import CurrencyRange
from tekmoney.currency_tax import CurrencyWithTax
from tekmoney.currency_range_tax import CurrencyRangeTax
from tekmoney.discount import fixed_discount, fractional_discount, percentage_discount


def test_application():
    price = CurrencyWithTax(Currency(30, 'BTC'), Currency(30, 'BTC'))
    discount = partial(fixed_discount, discount=Currency(10, 'BTC'))
    result = discount(price)
    assert result.net == Currency(20, 'BTC')
    assert result.gross == Currency(20, 'BTC')
    price_range = CurrencyRange(price.net, price.net)
    result = discount(price_range)
    assert result.start == Currency(20, 'BTC')
    assert result.stop == Currency(20, 'BTC')
    price_range = CurrencyRangeTax(price, price)
    result = discount(price_range)
    assert result.start == CurrencyWithTax(Currency(20, 'BTC'), Currency(20, 'BTC'))
    assert result.stop == CurrencyWithTax(Currency(20, 'BTC'), Currency(20, 'BTC'))
    with pytest.raises(TypeError):
        discount(1)


def test_zero_clipping():
    price = CurrencyWithTax(Currency(10, 'USD'), Currency(10, 'USD'))
    result = fixed_discount(price, Currency(30, 'USD'))
    assert result.net == Currency(0, 'USD')
    assert result.gross == Currency(0, 'USD')


def test_discount():
    price = CurrencyWithTax(Currency(100, 'BTC'), Currency(100, 'BTC'))
    discount = partial(fractional_discount, fraction=Decimal('0.25'))
    result = discount(price)
    assert result.net == Currency(75, 'BTC')
    assert result.gross == Currency(75, 'BTC')
    price_range = CurrencyRangeTax(price, price)
    result = discount(price_range)
    assert result.start == CurrencyWithTax(Currency(75, 'BTC'), Currency(75, 'BTC'))
    assert result.stop == CurrencyWithTax(Currency(75, 'BTC'), Currency(75, 'BTC'))
    result = discount(Currency(100, 'BTC'))
    assert result == Currency(75, 'BTC')
    with pytest.raises(TypeError):
        discount(100)


def test_discount_from_net():
    price = CurrencyWithTax(Currency(100, 'PLN'), Currency(200, 'PLN'))
    result = fractional_discount(price, Decimal('0.5'), from_gross=False)
    assert result.net == Currency(50, 'PLN')
    assert result.gross == Currency(150, 'PLN')


def test_currency_mismatch():
    with pytest.raises(ValueError):
        fixed_discount(
            CurrencyWithTax(Currency(10, 'BTC'), Currency(10, 'BTC')),
            Currency(10, 'USD'))


def test_percentage_discount():
    price = CurrencyWithTax(Currency(100, 'BTC'), Currency(100, 'BTC'))
    discount = partial(percentage_discount, percentage=10)
    result = discount(price)
    assert result.net == Currency(90, 'BTC')
    assert result.gross == Currency(90, 'BTC')
    price_range = CurrencyRangeTax(price, price)
    result = discount(price_range)
    assert result.start == CurrencyWithTax(Currency(90, 'BTC'), Currency(90, 'BTC'))
    assert result.stop == CurrencyWithTax(Currency(90, 'BTC'), Currency(90, 'BTC'))


def test_precision():
    price = CurrencyWithTax(Currency('1.01', 'BTC'), Currency('1.01', 'BTC'))
    result = percentage_discount(price, percentage=50)
    assert result.net == Currency('0.51', 'BTC')
    assert result.net == Currency('0.51', 'BTC')



from typing import Union
from .currency import Currency
from .currency_range import CurrencyRange
from .currency_tax import CurrencyWithTax

Addable = Union[Currency, CurrencyRange, CurrencyWithTax, 'CurrencyRangeTax']


class CurrencyRangeTax:
    """A taxed money range."""

    __slots__ = ('start', 'stop')

    def __init__(self, start: CurrencyWithTax, stop: CurrencyWithTax) -> None:
        if start.currency != stop.currency:
            raise ValueError(
                f'Cannot create a range. {start} and {stop} use different currencies'
            )
        if start > stop:
            raise ValueError(
                f'Cannot create a range from {start} to {stop}'
            )
        self.start = start
        self.stop = stop

    def __str__(self) -> str:
        return f'CurrencyRangeTax({self.start}, {self.stop})'

    def __add__(self, other: Addable) -> 'CurrencyRangeTax':
        if isinstance(other, (Currency, CurrencyWithTax)):
            if other.currency != self.currency:
                raise ValueError(
                    f'Cannot add a range in {self.currency} to argument in {other.currency}'
                )
            start = self.start + other
            stop = self.stop + other
            return CurrencyRangeTax(start, stop)
        elif isinstance(other, (CurrencyRange, CurrencyRangeTax)):
            if other.start.currency != self.currency:
                raise ValueError(
                    f'Cannot add ranges in {self.currency} and {other.currency}'
                )
            start = self.start + other.start
            stop = self.stop + other.stop
            return CurrencyRangeTax(start, stop)
        return NotImplemented

    def __sub__(self, other: Addable) -> 'CurrencyRangeTax':
        if isinstance(other, (Currency, CurrencyWithTax)):
            if other.currency != self.start.currency:
                raise ValueError(
                    f'Cannot subtract argument in {other.currency} from range in {self.currency}'
                )
            start = self.start - other
            stop = self.stop - other
            return CurrencyRangeTax(start, stop)
        elif isinstance(other, (CurrencyRange, CurrencyRangeTax)):
            if other.start.currency != self.start.currency:
                raise ValueError(
                    f'Cannot subtract range in {other.start.currency} from {self.currency}'
                )
            start = self.start - other.start
            stop = self.stop - other.stop
            return CurrencyRangeTax(start, stop)
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CurrencyRangeTax):
            return (
                self.start == other.start and
                self.stop == other.stop)
        return False

    def __contains__(self, item: CurrencyWithTax) -> bool:
        if not isinstance(item, CurrencyWithTax):
            raise TypeError(
                f'CurrencyRangeTax requires CurrencyWithTax as left operand, not {type(item)}'
            )
        return self.start <= item <= self.stop

    @property
    def currency(self) -> str:
        """Return the currency of the range."""
        return self.start.currency

    def quantize(self, exp=None, rounding=None) -> 'CurrencyRangeTax':
        """Return a copy of the range with start and stop quantized.

        All arguments are passed to `TaxedMoney.quantize` which in turn calls
        `Money.quantize`.
        """
        return CurrencyRangeTax(
            self.start.quantize(exp, rounding=rounding),
            self.stop.quantize(exp, rounding=rounding))

    def replace(self, start: CurrencyWithTax = None, stop: CurrencyWithTax = None) -> 'CurrencyRangeTax':
        """Return a range with start or stop replaced with given values."""
        if start is None:
            start = self.start
        if stop is None:
            stop = self.stop
        return CurrencyRangeTax(start=start, stop=stop)



import functools
import operator
from typing import Iterable, TypeVar

T = TypeVar('T')


def tek_sum(values: Iterable[T]) -> T:
    """ Return the sum of values in iterable"""
    return functools.reduce(operator.add, values)



import pytest
from decimal import Decimal
from tekmoney.currency import Currency


def test_str():
    currency = Currency(5, 'USD')
    assert str(currency) == '5 USD'


def test_bool():
    assert not Currency(0, 'USD')
    assert Currency(1, 'GBP')


def test_div():
    assert Currency(12, 'USD') / Currency(6, 'USD') == Decimal(2)
    assert Currency(12, 'USD') / Decimal('2') == Currency(6, 'USD')


def test_add():
    assert Currency(5, 'USD') + Currency(6, 'USD') == Currency(11, 'USD')
    assert not Currency(5, 'USD') + Currency(6, 'USD') == Currency(12, 'USD')
    with pytest.raises(ValueError):
        Currency(5, 'USD') + Currency(6, 'GBP')


def test_sub():
    assert Currency(6, 'USD') - Currency(6, 'USD') == Currency(0, 'USD')
    assert Currency(5, 'USD') - Currency(4, 'USD') == Currency(1, 'USD')
    assert not Currency(5, 'USD') - Currency(4, 'USD') == Currency(3, 'USD')
    with pytest.raises(ValueError):
        Currency(5, 'USD') - Currency(6, 'GBP')


def test_mul():
    assert Currency(6, 'USD') * 5 == Currency(30, 'USD')
    assert 5 * Currency(4, 'USD') == Currency(20, 'USD')
    assert not Currency(5, 'USD') * 8 == Currency(3, 'USD')


def test_lt():
    assert Currency(5, 'USD') < Currency(6, 'USD')
    assert not Currency(7, 'USD') < Currency(6, 'USD')
    with pytest.raises(ValueError):
        Currency(6, 'USD') < Currency(6, 'GBP')


def test_le():
    assert Currency(5, 'USD') <= Currency(5, 'USD')
    assert Currency(5, 'USD') <= Currency(6, 'USD')
    assert not Currency(7, 'USD') <= Currency(6, 'USD')
    with pytest.raises(ValueError):
        Currency(6, 'USD') <= Currency(6, 'GBP')


def test_gt():
    assert not Currency(5, 'USD') > Currency(6, 'USD')
    assert Currency(7, 'USD') > Currency(6, 'USD')
    with pytest.raises(ValueError):
        Currency(6, 'USD') > Currency(6, 'GBP')


def test_ge():
    assert Currency(5, 'USD') >= Currency(5, 'USD')
    assert not Currency(5, 'USD') >= Currency(6, 'USD')
    assert Currency(7, 'USD') >= Currency(6, 'USD')
    with pytest.raises(ValueError):
        Currency(6, 'USD') >= Currency(6, 'GBP')


def test_quantize():
    assert str(Currency(1, 'USD').quantize().amount) == '1.00'
    assert str(Currency(1, 'USD').quantize('.001').amount) == '1.000'
    assert str(Currency(Decimal(1.001), 'USD').quantize('.001').amount) == '1.001'
    assert str(Currency(Decimal(1.001), 'USD').quantize('.01').amount) == '1.00'




from setuptools import setup

with open('README.md', 'r') as f:
    long_description = f.read()
setup(
    name='tekmoney',         # How you named your package folder (MyLib)
    packages=['tekmoney'],   # Chose the same as "name"
    version='1.1',      # Start with a small number and increase it with every change you make
    license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='Handles money',   # Give a short description about your library
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='tek shinobi',                   # Type in your name
    author_email='hello@gmail.com',      # Type in your E-Mail
    url='https://github.com/tek-shinobi/tekmoney',   # Provide either the link to your github or to your website
    keywords=['money', 'currency'],   # Keywords that define your package best
    install_requires=[            # I get to this in a second
        'babel>=2.5.0',
        'typing>=3.6.0',
    ],
    platforms=['any'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',      # Define that your audience are developers
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        'Programming Language :: Python :: 3',      # Specify which python versions that you want to support
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
    ]
)
